/*
 * Copyright 2010-2012 the original author or authors.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.callbackparams.bdd.support;

import java.lang.annotation.Annotation;
import java.lang.reflect.Field;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.math.BigInteger;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.LinkedHashSet;
import java.util.List;
import org.callbackparams.ext.CallbackControlPanel;
import org.callbackparams.internal.template.CallbackRecord;
import org.callbackparams.internal.template.TestrunCallbacks;
import org.callbackparams.bdd.BddOrder;

/**
 * @author Henrik Kaipe
 */
public class GenericBddRunner {

    private TestrunCallbacks testInstance;
    private CallbackControlPanel ccp;
    private Collection<Field> testFields = new ArrayList<Field>();
    private List<Method> methodsInOrder = new ArrayList<Method>();

    private GenericBddRunner(TestrunCallbacks testInstance) {
        this.testInstance = testInstance;
    }

    public static void run(final TestrunCallbacks testInstance)
    throws Throwable {
        GenericBddRunner gbr = new GenericBddRunner(testInstance);
        gbr.init();
        gbr.execute();
    }

    private void execute() throws Throwable {
        for (Method m : methodsInOrder) {
            for (Object o : potentialInvokeTargets()) {
                if (m.getDeclaringClass().isAssignableFrom(o.getClass())) {
                    try {
                        m.invoke(o, pickArguments(m.getParameterTypes()));
                    } catch (InvocationTargetException ite) {
                        throw ite.getTargetException();
                    }
                }
            }
        }
    }

    private Iterable<Object> potentialInvokeTargets() {
        List record = ccp.getCurrentCallbackRecord();
        List targets = new ArrayList(record.size() + 1);
        targets.add(testInstance);
        targets.addAll(record);
        return targets;
    }

    private Object[] pickArguments(Class<?>[] paramTypes)
    throws IllegalAccessException {
        Object[] args = new Object[paramTypes.length];
        for (int i = 0 ; i < paramTypes.length ; ++i) {
            args[i] = pickArgument(paramTypes[i]);
        }
        return args;
    }

    private Object pickArgument(Class<?> type) throws IllegalAccessException {
        if (type.isInstance(testInstance)) {
            return testInstance;
        } else {
            for (Field f : testFields) {
                if (type.isAssignableFrom(f.getType())) {
                    return f.get(testInstance);
                }
            }
        }
        return null;
    }

    private void init() {
        ccp = getCallbackRecord().getCallbackControlPanel();
        collectTestField();
        collectMethodsInOrder();
    }

    private CallbackRecord getCallbackRecord() {
        try {
            Method m = TestrunCallbacks.class
                    .getDeclaredMethod("getCallbackRecord");
            m.setAccessible(true);
            return (CallbackRecord) m.invoke(testInstance);
        } catch (InvocationTargetException ex) {
            throw new Error(ex.getTargetException());
        } catch (Exception ex) {
            throw new Error(ex);
        }
    }

    private void collectMethodsInOrder() {
        Collection<Method> findOrderMethods = new LinkedHashSet<Method>();

        for (Class<?> c : testClassHierarchy()) {
            for (Method m : c.getDeclaredMethods()) {
                if (isBddMethod(m)) {
                    m.setAccessible(true);
                    findOrderMethods.add(m);
                }
            }
        }

        for (Object v : recordValues()) {
            for (Method m : v.getClass().getDeclaredMethods()) {
                if (isBddMethod(m)) {
                    m.setAccessible(true);
                    findOrderMethods.add(m);
                }
            }
        }

        methodsInOrder.addAll(findOrderMethods);

        Collections.sort(methodsInOrder, new Comparator<Method>() {

            public int compare(Method m1, Method m2) {
                return order(m1).compareTo(order(m2));
            }

            private BigInteger order(Method m) {
                BigInteger order = BigInteger.ZERO;
                for (Annotation a : m.getAnnotations()) {
                    Class<?> annotationClass = a.annotationType();
                    if (annotationClass.isAnnotationPresent(BddOrder.class)) {
                        order = order.add(BigInteger.valueOf(annotationClass
                                .getAnnotation(BddOrder.class).value()));
                    }
                }
                return order;
            }
        });
    }

    private boolean isBddMethod(Method m) {
        for (Annotation a : m.getAnnotations()) {
            if (a.annotationType().isAnnotationPresent(BddOrder.class)) {
                return true;
            }
        }
        return false;
    }

    private void collectTestField() {
        for (Class<?> c : testClassHierarchy()) {
            for (Field f : c.getDeclaredFields()) {
                f.setAccessible(true);
                testFields.add(f);
            }
        }
    }

    private Iterable<Object> recordValues() {
        return ccp.getCurrentCallbackRecord();
    }

    private Iterable<Class<?>> testClassHierarchy() {
        List classes = new ArrayList();
        for (Class<?> c = testInstance.getClass()
                ; TestrunCallbacks.class != c ; c = c.getSuperclass()) {
            classes.add(c);
        }
        return classes;
    }
}
