/*
 * Copyright 2010-2012 the original author or authors.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.callbackparams.internal;

import java.io.IOException;
import java.io.InputStream;
import java.net.URL;
import java.util.Collection;
import java.util.Collections;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.NoSuchElementException;
import java.util.Set;
import org.apache.bcel.classfile.JavaClass;
import org.apache.bcel.util.ClassLoaderRepository;
import org.callbackparams.support.ResourceMap;

/**
 * @author Henrik Kaipe
 */
public class CallbackTestClassReloader extends ClassLoader {

    public interface ReloadFilter {
        boolean needsToBeReloaded(String classNameToReload);
    }

    interface Rebyter {

        /**
         * Performs custom byte-code modifications of the class about to be
         * reloaded. But whatever you do - you must not change the class-name!!
         *
         * @param originalClassDef class-definition as retrieved from the
         *        original classloader
         * @return the byte array with the new class-definition's byte-code
         */
        byte[] rebyte(
                CallbackContext callbackContext,
                JavaClass templateClassDef);

        byte[] newSyntheticClass(String className);
    }

    private final ClassLoader templateClassLoader;
    private final ClassLoaderRepository templateClassLoaderRepo;
    private final ReloadFilter reloadFilter;
    private final Rebyter rebyter;

    private final Set redefinedPackages = new HashSet();
    private final Map reloadedClasses = new HashMap();
    private final ResourceMap reloadedResources = new ResourceMap();

    private CallbackContextQue callbackContextQue =
            new CallbackContextQue(null); // Will be replace by construcor

    private CallbackTestClassReloader(ClassLoader parentClassLoader,
            ClassLoader templateClassLoader,
            Rebyter rebyter, ReloadFilter reloadFilter) {
        super(parentClassLoader);
        setDefaultAssertionStatus(true);

        this.templateClassLoader = templateClassLoader;
        this.reloadFilter = reloadFilter;
        this.rebyter = rebyter;
        this.templateClassLoaderRepo =
                new ClassLoaderRepository(templateClassLoader);
        try {
            this.callbackContextQue = new CallbackContextQue(
                    CallbackContextFactory.forClassLoader(this));
        } catch (Exception ex) {
            if (ex instanceof RuntimeException) {
                throw (RuntimeException)ex;
            } else {
                throw new Error(ex);
            }
        }
    }

    public static CallbackTestClassReloader newInstance(Class testClass,
            ReloadFilter specialReloadFilter, CallbackRebyteInfo rebyteInfo) {
        CallbackMethodProxyingRebyter rebyter =
                new CallbackMethodProxyingRebyter(testClass, rebyteInfo);
        return new CallbackTestClassReloader(
                rebyter.getParentClassLoader(),
                testClass.getClassLoader(),
                rebyter,
                new GeneralReloadFilter(specialReloadFilter));
    }

    //    @Override cannot be used under JDK-1.4.x
    public URL getResource(String name) {
        URL rebytedResource = reloadedResources.getURL(name);
        if (null != rebytedResource) {
            return rebytedResource;
        } else {
            return templateClassLoader.getResource(name);
        }
    }

//    @Override cannot be used under JDK-1.4.x
    public InputStream getResourceAsStream(String name) {
        InputStream is = reloadedResources.getAsStream(name);
        if (null != is) {
            return is;
        } else {
            return templateClassLoader.getResourceAsStream(name);
        }
    }

//    @Override cannot be used under JDK-1.4.x
    protected Class findClass(String name) throws ClassNotFoundException {
        if (reloadFilter.needsToBeReloaded(name)) {
            return findReloadedClass(name);

        } else {
            return templateClassLoader.loadClass(name);
        }
    }

    private void ensurePackageIsRedefined(String packageName)
    throws IllegalArgumentException {
        if (redefinedPackages.add(packageName)) {
            definePackage(packageName,
                    null, null, null, null, null, null, null);
        }
    }

    private Class findReloadedClass(String name) throws ClassNotFoundException {
        if (false == reloadedClasses.containsKey(name)) {
            try {
                ensurePackageIsRedefined(name
                        .substring(0, Math.max(0, name.lastIndexOf('.'))));
            } catch (IllegalArgumentException usuallyMeansANonExistentApiClass){
                System.err.println("Cannot redefine package for class " + name);
            }
            reloadClass(name);
        }
        return (Class) reloadedClasses.get(name);
    }

    private void reloadClass(String name) throws ClassNotFoundException {
        byte[] byteCode = rebyter.newSyntheticClass(name);
        if (null == byteCode) {
            byteCode = rebyter.rebyte(callbackContextQue,
                    templateClassLoaderRepo.loadClass(name));
        }
        Class reloadedClass = defineClass(name, byteCode, 0, byteCode.length);
        reloadedResources.put(reloadedClass, byteCode);
        reloadedClasses.put(name, reloadedClass);
    }

    public void setupTestRunCallbackRecord(List callbackRecord) {
        getCallbackContext().setCallbackRecord(callbackRecord);
    }

    public CallbackContext getCallbackContext() {
        this.callbackContextQue.releaseQue();
        return this.callbackContextQue.getTargetContext();
    }

    //    @Override cannot be used under JDK-1.4.x
    /**
     * Handles the necessary stuff for making method
     * {@link getResources(String)} work properly.
     * Please note that in JDK-1.4 the method {@link getResources(String)}
     * is final! - That is why this method was overridden instead.
     */
    protected Enumeration findResources(String name) throws IOException {
        final URL url = reloadedResources.getURL(name);

        if (null != url) {
            return singletonEnumeration(url);

        } else {
            return findTemplateClassLoaderResources(name);
        }
    }

    private Enumeration singletonEnumeration(final URL singleElement) {
        return new Enumeration() {
            boolean elementAvailable = true;

            public boolean hasMoreElements() {
                return elementAvailable;
            }

            public Object nextElement() {
                if (hasMoreElements()) {
                    elementAvailable = false;
                    return singleElement;
                } else {
                    throw new NoSuchElementException(
                            "No more resource available");
                }
            }
        };
    }

    private Enumeration findTemplateClassLoaderResources(String name)
    throws IOException {
        Enumeration templateResources = templateClassLoader.getResources(name);

        /* Possible duplicate urls from parent should be removed ... */
        if (null != getParent()) {
            Collection parentResources = Collections
                    .list(getParent().getResources(name));

            while (templateResources.hasMoreElements()
                    && false == parentResources.isEmpty()) {
                URL duplicate = (URL) templateResources.nextElement();

                if (false == parentResources.remove(duplicate)) {
                    /* Weird situation - seems like the template
                     * class-loader somehow censurs the resources of its
                     * parent. Best way to resolve the situation is
                     * probably to return all resources available through
                     * the template class-loader ... */
                    return templateClassLoader.getResources(name);
                }
            }
        }

        return templateResources;
    }
}
