/*
 * Copyright 2010-2013 the original author or authors.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.callbackparams.internal.template;

import java.lang.reflect.Constructor;
import java.lang.reflect.Field;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.ListIterator;
import org.callbackparams.factory.CallbackFactory;

/**
 * @author Henrik Kaipe
 */
public abstract class CallbackRef {

    private final Class paramClass;
    private final AnnotationSpecifiedCallbacks annotationSpecified;
    private final CallbackFactory secondaryFactory;

    public static CallbackRef newInstance(Method m, int paramIndex) {
        return new CallbackRef(m, paramIndex) {};
    }

    static CallbackRef newInstance(Field f) {
        return new CallbackRef(f) {};
    }

    interface CallbackFactoryFinder {
        CallbackFactory find(Class interf);
    }

    /**
     * Search for a field that is declared as CallbackFactory ...
     */
    private static final CallbackFactoryFinder FACTORY_FIELD_FINDER =
            new CallbackFactoryFinder() {
        public CallbackFactory find(Class interf) {

            final Field[] fields = interf.getDeclaredFields();
            for (int i = 0 ; i < fields.length ; ++i) {
                final Field f = fields[i];

                try {
                    f.setAccessible(true);
                } catch (SecurityException ignore) {
                    /* Let's try to continue anyway ... */
                }

                if (CallbackFactory.class.isAssignableFrom(f.getType())) {
                    try {
                        return (CallbackFactory) f.get(null);
                    } catch (Exception x) {
                        throw new Error(x);
                    }
                }
            }
            return null;
        }
    };

    private static CallbackFactoryFinder findByInnerClassConstructor(
            final Class[] constructorParamTypes,
            final Object[] constructorArgs) {

        return new CallbackFactoryFinder() {
            public CallbackFactory find(Class interf) {

                final Class[] nestedClasses = interf.getDeclaredClasses();
                for (int i = 0 ; i < nestedClasses.length ; ++i) {
                    final Class nested = nestedClasses[i];
                    if (false == CallbackFactory.class.isAssignableFrom(nested)
                            || Modifier.isAbstract(nested.getModifiers())) {
                        continue;
                    }

                    Constructor constructor = null;
                    try {
                        constructor = nested
                            .getDeclaredConstructor(constructorParamTypes);
                        constructor.setAccessible(true);

                    } catch (NoSuchMethodException x) {
                        continue;
                    } catch (SecurityException ignore) {
                        /* Let's hope it will work anyway ... */
                    }

                    try {
                        return (CallbackFactory) constructor
                                .newInstance(constructorArgs);

                    } catch (Exception noGood) {
                        Throwable toThrow;
                        if (noGood instanceof InvocationTargetException) {
                            toThrow = ((InvocationTargetException)noGood)
                                    .getTargetException();
                        } else {
                            toThrow = noGood;
                        }
                        if (toThrow instanceof RuntimeException) {
                            throw (RuntimeException)toThrow;
                        } else {
                            throw new Error(toThrow);
                        }
                    }
                }
                return null;
            }
        };
    }

    private static final CallbackFactory NULL_FACTORY = new CallbackFactory() {
        public Object asCallback(Object callbackRecordElement) {
            return null;
        }
    };

    private static Iterator interfaces(Class interf) {
        final List interfaces = new ArrayList();
        interfaces.add(interf);

        return new Iterator() {
            Iterator iter = interfaces.iterator();

            public boolean hasNext() {
                return false == interfaces.isEmpty();
            }
            public Object next() {
                try {
                    return iter.next();
                } finally {
                    if (false == iter.hasNext()) {
                        refreshInterfacesList();
                        iter = interfaces.iterator();
                    }
                }
            }
            public void remove() {
                throw new UnsupportedOperationException("Not supported yet.");
            }

            void refreshInterfacesList() {
                ListIterator li = interfaces.listIterator();
                while (li.hasNext()) {
                    final Class old = (Class) li.next();
                    li.remove();
                    Class[] superInterfaces = old.getInterfaces();
                    for (int i = 0; i < superInterfaces.length; ++i) {
                        li.add(superInterfaces[i]);
                    }
                }
            }
        };
    }

    private CallbackFactory lookupDeclaredCallbackFactory(Field f) {
        return lookupDeclaredCallbackFactory(findByInnerClassConstructor(
                new Class[] {Field.class}, new Object[] {f}));
    }

    private CallbackFactory lookupDeclaredCallbackFactory(
            Method m, int paramIndex) {
        return lookupDeclaredCallbackFactory(findByInnerClassConstructor(
                new Class[] {Method.class, int.class},
                new Object[] {m, new Integer(paramIndex)}));
    }

    private CallbackFactory lookupDeclaredCallbackFactory(
            CallbackFactoryFinder finderBackup) {
        CallbackFactoryFinder[] finders = {
            FACTORY_FIELD_FINDER, finderBackup
        };

        for (Iterator iter = interfaces(paramClass) ; iter.hasNext() ;) {
            final Class interf = (Class) iter.next();

            for (int i = 0 ; i < finders.length ; ++i) {
                final CallbackFactoryFinder factoryFinder = finders[i];

                CallbackFactory factory = factoryFinder.find(interf);
                if (null != factory) {
                    return factory;
                }
            }
        }

        return NULL_FACTORY;
    }

    CallbackRef(Method m, int paramIndex) {
        this.paramClass = m.getParameterTypes()[paramIndex];
        possibleDeprecationWarning(m);
        this.annotationSpecified =
                AnnotationSpecifiedCallbacks.forCallbackRef(m, paramIndex);
        this.secondaryFactory = lookupDeclaredCallbackFactory(m, paramIndex);
    }

    CallbackRef(Field f) {
        this.paramClass = f.getType();
        possibleDeprecationWarning(f);
        this.annotationSpecified =
                AnnotationSpecifiedCallbacks.forCallbackRef(f);
        this.secondaryFactory = lookupDeclaredCallbackFactory(f);
    }

    private void possibleDeprecationWarning(java.lang.reflect.Member m) {
        if (org.callbackparams.CallbackControlPanel.class == paramClass) {
            System.err.println(m + org.callbackparams.CallbackControlPanel
                    .DEPRECATION_MESSAGE);
        }
        CallbackFieldWarning.possibleDeprecationWarning(m);
    }

    Object asCallback(Object callbackRecordElement) {

        if (isCompatibleCallback(callbackRecordElement)) {
            return callbackRecordElement;
        }

        if (callbackRecordElement instanceof org.callbackparams.CallbackFactory) {
            Object suppliedCallback =
                    ((org.callbackparams.CallbackFactory)callbackRecordElement).getCallback();
            if (isCompatibleCallback(suppliedCallback)) {
                return suppliedCallback;
            }
        }

        Object callback = annotationSpecified.asCallback(callbackRecordElement);
        if (null != callback) {
            return callback;
        }

        /* Using secondary CallbackFactory is the last resort ... */
        return this.secondaryFactory.asCallback(callbackRecordElement);
    }

    boolean isCompatibleCallback(Object possibleCallback) {
        return this.paramClass.isInstance(possibleCallback);
    }

    Class getParamClass() {
        return paramClass;
    }
}
