/*
 * Copyright 2013 the original author or authors.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.callbackparams.internal.template;

import java.lang.reflect.Method;
import java.util.IdentityHashMap;
import java.util.Map;
import org.callbackparams.support.ClassWrapper;
import org.callbackparams.support.DefaultValue;
import org.callbackparams.support.JdkVersion;

/**
 * Digests all callback return-values during the overall callback-invocation and
 * chooses what value to return from the overall callback-invocation.
 * Also keeps track of which arguments to pass to the callback.
 *
 * @author Henrik Kaipe
 */
public class ReturnValueDigester {

    private static final ClassWrapper implClass =
            JdkVersion.compatibleImplementationOf(ReturnValueDigester.class);

    private static final Object
            NO_RETURN_VALUE_HAS_BEEN_DIGESTED_YET = new Object();

    private Class returnType;
    private boolean returnTypeAndFirstArgumentTypeAreSame;
    private Object[] args;
    private Object currentReturnValue = NO_RETURN_VALUE_HAS_BEEN_DIGESTED_YET;
    private Map returnValuesSoFar = new IdentityHashMap();

    static ReturnValueDigester newInstance(Method m, Object[] args) {
        ReturnValueDigester instance =
                (ReturnValueDigester) implClass.newInstance();
        instance.init(m, args);
        return instance;
    }

    /**
     * Not for direct use - use factory-method
     * {@link #newInstance(java.lang.reflect.Method, java.lang.Object[])}
     * instead!
     */
    protected ReturnValueDigester() {}

    private void init(Method m, Object[] args) {
        this.returnType = m.getReturnType();
        this.returnTypeAndFirstArgumentTypeAreSame =
                returnTypeAndFirstArgumentTypeAreSame(m);
        this.args = null == args ? null : (Object[]) args.clone();
        this.currentReturnValue = this.returnTypeAndFirstArgumentTypeAreSame
                ? args[0]
                : NO_RETURN_VALUE_HAS_BEEN_DIGESTED_YET;
    }

    void digestNext(Object callbackRecordElement, Object returnValue) {
        returnValuesSoFar.put(callbackRecordElement, returnValue);

        if (this.returnTypeAndFirstArgumentTypeAreSame) {
            this.args[0] = this.currentReturnValue = returnValue;

        } else if (NO_RETURN_VALUE_HAS_BEEN_DIGESTED_YET == this.currentReturnValue
                || null == returnValue
                || returnValue.equals(this.currentReturnValue)) {
            this.currentReturnValue = returnValue;

        } else {
            this.currentReturnValue = DefaultValue.forType(this.returnType);
        }
    }

    void updateLatestCallbackResultsInCallbackControlPanel(Map latestResults) {
        latestResults.clear();
        latestResults.putAll(returnValuesSoFar);
    }

    Object returnValue() {
        return NO_RETURN_VALUE_HAS_BEEN_DIGESTED_YET == currentReturnValue
                ? DefaultValue.forType(this.returnType)
                : this.currentReturnValue;
    }

    protected boolean returnTypeAndFirstArgumentTypeAreSame(Method m) {
        Class[] methodParams = m.getParameterTypes();
        return 1 <= methodParams.length
                && m.getReturnType() == methodParams[0];
    }

    Object[] getArguments() {
        return this.args;
    }
}
