/*
 * Copyright 2012 the original author or authors.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.callbackparams.internal.template.annotation;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import org.callbackparams.CallbackFactory;
import org.callbackparams.ext.WrappingCallbackImpls;
import org.callbackparams.wrap.legacy.Wrapping;

/**
 * @author Henrik Kaipe
 */
class WrappingCreator<T> {

    private final Class<T> callbackType;
    private final ValidatingWrappingFactory<T> factory;
    private final List<Class<? extends Wrapping>> validSubClasses =
            new ArrayList<Class<? extends Wrapping>>();

    WrappingCreator(
            Class<T> callbackType,
            Object value,
            Class<? extends Wrapping>[] subClasses) {

        this.callbackType = callbackType;
        this.factory = new ValidatingWrappingFactory<T>(value);
        for (Class<? extends Wrapping> sub : subClasses) {
            if (this.factory.isCompatibleWrappedValue(sub)
                    && (callbackType.isAssignableFrom(sub)
                    || CallbackFactory.class.isAssignableFrom(sub))) {
                validSubClasses.add(sub);
            }
        }
    }

    T createCallbackInstance(
            List<Class<? extends Wrapping>[]> wrappingImplOptions,
            boolean filterOnSubclasses) {

        LinkedList<LinkedList<Class<? extends Wrapping>>> callbackFactoryQues =
                new LinkedList<LinkedList<Class<? extends Wrapping>>>();

        Class<? extends Wrapping> pendingWrappingImplClass =
                setupCallbackFactoryQues(callbackFactoryQues, wrappingImplOptions);
        T pendingCallbackInstance = null;
        while (false == callbackFactoryQues.isEmpty()) {
            for (Iterator<LinkedList<Class<? extends Wrapping>>> iterFactoryQues =
                    callbackFactoryQues.iterator(); iterFactoryQues.hasNext();) {
                final Class<? extends Wrapping>
                        factoryWrappingImpl = iterFactoryQues.next().get(0);
                if (factoryWrappingImpl
                        .isAssignableFrom(pendingWrappingImplClass)) {
                    iterFactoryQues.remove();
                    continue;
                }
                T possibleCallbackInstance =
                        factory.newInstance(factoryWrappingImpl);
                if (false == callbackType.isInstance(possibleCallbackInstance)) {
                    continue;

                } else if (pendingWrappingImplClass
                            .isAssignableFrom(factoryWrappingImpl)) {
                    pendingWrappingImplClass = factoryWrappingImpl;
                    pendingCallbackInstance = possibleCallbackInstance;
                    iterFactoryQues.remove();
                    continue;

                } else {
                    throw new Error("Multiple implementations of callback-"
                            + callbackType
                            + " have been specified by @"
                            + WrappingCallbackImpls.class.getName()
                            + "-annotated annotations! Please use an explicit @"
                            + WrappingCallbackImpls.class.getName()
                            + "-annotation on your class or enum-constant to make a stringent specification of the your prefered "
                            + callbackType.getName() + "-implementation!!");
                }
            }
            for (Iterator<LinkedList<Class<? extends Wrapping>>> iterFactoryQues =
                    callbackFactoryQues.iterator(); iterFactoryQues.hasNext();) {
                final LinkedList<Class<? extends Wrapping>> pendingQue =
                        iterFactoryQues.next();
                if (1 == pendingQue.size() || pendingQue.removeFirst()
                        .isAssignableFrom(pendingWrappingImplClass)) {
                    iterFactoryQues.remove();
                }
            }
        }

        if (Wrapping.class == pendingWrappingImplClass) {
            /* None of the wrappings work here ... */
            return null;

        } else if (filterOnSubclasses) {
            for (Class<? extends Wrapping> sub : validSubClasses) {
                if (sub != pendingWrappingImplClass
                        && pendingWrappingImplClass.isAssignableFrom(sub)) {
                    if (CallbackFactory.class.isAssignableFrom(sub)) {
                        T possibleCallbackInstance = factory.newInstance(sub);
                        if (callbackType.isInstance(possibleCallbackInstance)) {
                            pendingWrappingImplClass = sub;
                            pendingCallbackInstance = possibleCallbackInstance;
                        }
                    } else {
                        pendingWrappingImplClass = sub;
                    }
                }
            }
        }
        return null != pendingCallbackInstance ? pendingCallbackInstance
                : factory.newInstance(pendingWrappingImplClass);
    }

    Class<? extends Wrapping> setupCallbackFactoryQues(
            List<LinkedList<Class<? extends Wrapping>>> factoryQues,
            List<Class<? extends Wrapping>[]> wrappingImplOptions) {
        Class<? extends Wrapping> callbackImplClass = Wrapping.class;

        for (Class[] wrappingImpls : wrappingImplOptions) {
            LinkedList<Class<? extends Wrapping>> currentQue = null;

            for (Class wrappingImpl : wrappingImpls) {

                if (false == factory.isCompatibleWrappedValue(wrappingImpl)) {
                    continue;

                } else if (CallbackFactory.class.isAssignableFrom(wrappingImpl)) {
                    if (null == currentQue) {
                        factoryQues.add(currentQue =
                                new LinkedList<Class<? extends Wrapping>>());
                    }
                    currentQue.add(wrappingImpl);

                } else if (null == currentQue
                        && callbackType.isAssignableFrom(wrappingImpl)) {
                    if (callbackImplClass.isAssignableFrom(wrappingImpl)) {
                        callbackImplClass = wrappingImpl;

                    } else if (false == wrappingImpl
                            .isAssignableFrom(callbackImplClass)) {
                        throw new Error("Multiple implementations of callback-"
                                + callbackType
                                + " have been specified by @"
                                + WrappingCallbackImpls.class.getName()
                                + "-annotated annotations! Please use an explicit @"
                                + WrappingCallbackImpls.class.getName()
                                + "-annotation on your class or enum-constant to make a stringent specification of the prefered "
                                + callbackType.getName() + "-implementation!!");
                    }
                    break;
                }
            }
        }
        return callbackImplClass;
    }
}
