/*
 * Copyright 2010-2013 the original author or authors.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.callbackparams.junit3;

import org.callbackparams.junit3.support.RebyteInfo;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import junit.framework.TestSuite;
import org.callbackparams.combine.reflect.CallbackRecordsFactory;
import org.callbackparams.internal.CallbackTestClassReloader;

/**
 * TestSuite implementation that could be useful under JUnit-3.8.x when
 * CallbackTestCase cannot be subclassed, such as when it is desired to
 * subclass another third-party TestCase implementation, e.g. MockObjectTestCase
 * in JMock-1.x.
 * <br/><br/>
 * Example usage ...
 * <pre>
 * public class MyTest extends MockObjectTestCase {
 *
 *     private CallbackInterface foo;
 * 
 *     //
 *     // ... Test-logic and param-values goes here ...
 *     //
 *
 *     public static Test suite() {
 *         return new CallbackTestSuite(MyTest.class);
 *     }
 * }
 * </pre>
 * To get an idea of how the above actually works under JUnit-3.8.x
 * <a target="_blank"
 * href="http://users.csc.calpoly.edu/~jdalbey/308/Resources/JUnitPrimer.html#testsuite"
 * >this article</a> might shed some light.
 * <br/><br/>
 * Upgrade to JUnit-4.x and usage of
 * {@link org.callbackparams.junit4.CallbackParamsRunner CallbackParamsRunner}
 * is strongly recommended! Thereafter the above test-example should
 * be refactored to ...
 * <pre>
 * &#64;RunWith(CallbackParamsRunner.class)
 * public class MyTest extends MockObjectTestCase {
 *
 *     {@link org.callbackparams.ParameterizedCallback @ParameterizedCallback}
 *     private CallbackInterface foo;
 *
 *     //
 *     // ... Test-logic and param-values goes here ...
 *     //
 * }
 * </pre>
 * I.e. method <code>suite()</code> is replaced by the class-annotation
 * <code>&#64;RunWith(CallbackParamsRunner.class)</code> and the
 * callback-injected field <code>foo</code> is annotated with
 * {@link org.callbackparams.ParameterizedCallback @ParameterizedCallback}.
 * (Trying to use <code>RunWith(CallbackParamsRunner.class)</code> while
 * still having <code>suite()</code> return a <code>CallbackTestSuite</code>
 * instance will result in undefined incorrect behaviour.)
 *
 * @author Henrik Kaipe
 */
public class CallbackTestSuite extends TestSuite {

    /**
     * Special constructor that allows the user to specify the
     * {@link CallbackRecordsFactory} instance. This API is pretty much
     * CallbackTestSuite's equivalent for
     * {@link CallbackTestCase#getCallbackRecordsFactory()}.
     *
     * @see CallbackRecordsFactory
     */
    public CallbackTestSuite(
            final Class testClass, final CallbackRecordsFactory factory) {
        Map/*String,String*/ newTestNamesMap = new HashMap();
        Class rebytedTestClass = rebyteTestClass(testClass, newTestNamesMap);

        TestInstanceFactory tif = new TestInstanceFactory(rebytedTestClass);
        for (Iterator iter = factory
                .collectCallbackRecordsReflectively(rebytedTestClass).iterator()
                ; iter.hasNext() ;) {
            final Object record = iter.next();

            final Method[] publicMethods = rebytedTestClass.getMethods();
            for (int i = 0 ; i < publicMethods.length ; ++i) {
                final Method m = publicMethods[i];
                if (void.class != m.getReturnType()
                        || Modifier.isStatic(m.getModifiers())
                        || 0 != m.getParameterTypes().length
                        || false == m.getName().startsWith("test")) {
                    continue;
                }
                String noargTestMethodName = m.getName();
                String actualTestMethodName =
                        newTestNamesMap.containsKey(noargTestMethodName)
                        ? (String) newTestNamesMap.get(noargTestMethodName)
                        : noargTestMethodName;
                addTest(tif.newTestrun(
                        noargTestMethodName, actualTestMethodName, record));
            }
        }
    }

    public CallbackTestSuite(final Class testClass) {
        this(testClass, CallbackRecordsFactory.getInstance());
    }

    static Class rebyteTestClass(
            final Class testClass, final Map newTestNamesMap) {
        RebyteInfo crInfo = new RebyteInfo() {
            public boolean isCallbackProxiedMethod(Method m) {
                return null != newTestNamesMap
                        && super.isCallbackProxiedMethod(m);
            }
            public void putNoargProxyMethod(
                    String proxyMethodName, Method callbackProxiedMethod) {
                newTestNamesMap.put(proxyMethodName,
                        callbackProxiedMethod.getName());
            }
        };
        CallbackTestClassReloader classReloader = CallbackTestClassReloader
                .newInstance(testClass, null, crInfo);
        try {
            Class reloadedClass = classReloader.loadClass(testClass.getName());
            classReloader.getCallbackContext();
            return reloadedClass;
        } catch (ClassNotFoundException x) {
            throw new Error(x);
        }
    }
}
