/*
 * Copyright 2010-2011 the original author or authors.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.callbackparams.junit4.version;

import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import org.callbackparams.junit4.EnumRunner;
import org.junit.runner.RunWith;
import org.junit.runner.Runner;
import org.junit.runner.notification.RunNotifier;

/**
 * Works as factory for creating instances whose implementation depends on the
 * JUnit-version currently in use.
 *
 * @author Henrik Kaipe
 */
public class Version {

    private enum JUnitVersion {
        JUNIT_4_5_OR_LATER, JUNIT_4_4, JUNIT_4_3_OR_EARLIER;
    }

    /* The initial value is supposed to be the best but will be changed in
     * case it does not work properly */
    private static JUnitVersion junitVersion = JUnitVersion.JUNIT_4_5_OR_LATER;

    private static final Constructor<? extends Runner> junit4Runner = runConstr(
            "org.junit.runners.JUnit4",
            JUnitVersion.JUNIT_4_4,
            "org.junit.internal.runners.TestClassRunner");
    private static final Constructor<? extends Runner> junit38Runner= runConstr(
            "org.junit.internal.runners.JUnit38ClassRunner",
            JUnitVersion.JUNIT_4_3_OR_EARLIER,
            "org.junit.internal.runners.OldTestClassRunner");
    private static final Constructor<? extends Runner> junit38SuiteRunner =
            runConstr("org.junit.runners.AllTests", null, null);

    private static final Constructor<? extends Runner> enumRunner =
            junitVersion != JUnitVersion.JUNIT_4_5_OR_LATER
            ? null
            : runConstr(EnumRunner.class.getName(), null, null);

    private static Constructor<? extends Runner> runConstr(String desiredClass,
            JUnitVersion backupVersion, String backupClass) {
        try {
            try {
                return Class.forName(desiredClass).asSubclass(Runner.class)
                        .getConstructor(Class.class);
            } catch (ClassNotFoundException ex) {
                junitVersion = backupVersion;
                try {
                    return Class.forName(backupClass).asSubclass(Runner.class)
                            .getConstructor(Class.class);
                } catch (ClassNotFoundException ex1) {
                    throw new Error(ex1);
                }
            }
        } catch (NoSuchMethodException ex) {
            throw new Error(ex);
        }
    }

    private static Runner newRunner(
            Constructor<? extends Runner> constr, Class<?> testClass) {
        try {
            try {
                return constr.newInstance(testClass);
            } catch (InvocationTargetException x) {
                throw x.getTargetException();
            }
        } catch (Throwable x) {
            if (x instanceof RuntimeException) {
                throw (RuntimeException)x;
            } else if (x instanceof Error) {
                throw (Error)x;
            } else {
                throw new Error(x);
            }
        }
    }

    public static Runner newJunit4Runner(Class<?> testClass) {
        return newRunner(junit4Runner, testClass);
    }

    public static Runner newJunit38Runner(Class<?> testClass) {
        return newRunner(junit38Runner, testClass);
    }

    public static Runner newJunit38SuiteRunner(Class<?> testClass) {
        return newRunner(junit38SuiteRunner, testClass);
    }

    public static Runner newEnumRunner(Class<? extends Enum> testClass) {
        switch (junitVersion) {
            case JUNIT_4_5_OR_LATER:
                return newRunner(enumRunner, testClass);
            default:
                String runnerName = testClass
                        .getAnnotation(RunWith.class).value().getName();
                throw new Error(
                        runnerName.substring(runnerName.lastIndexOf('.') + 1) +
                        " needs JUnit-4.5 or later to run an enum test-class");
        }
    }

    public static CallbackRunListener newRunListener(RunNotifier notifier) {
        switch (junitVersion) {
            default:
                return new CallbackRunListener_4_4_OrEarlier(notifier);
            case JUNIT_4_5_OR_LATER:
                return new CallbackRunListener_4_5(notifier);
        }
    }
}
