/*
 * Copyright 2013 the original author or authors.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.callbackparams.support;

import java.util.regex.Pattern;

/**
 * Provides utility methods to help out wherever the framework shall
 * act in different ways on different JDK versions.
 * The only thing that is checked on the JDK version is whether it supports
 * Java5-or-later (i.e. are annotations supported) or if only Java 1.4.x
 * is supported.
 *
 * @author Henrik Kaipe
 */
public class JdkVersion {

    private static final boolean annotationsAreSupported = new Object() {
        public int hashCode() {
            try {
                Class.forName("java.lang.annotation.Annotation");
                return 5;
            } catch (Exception detection_of_JDK_version_1_4_or_earlier) {
                return 4;
            }
        }
    }.hashCode() >= 5;
    private static final Pattern packageName = annotationsAreSupported
            ? Pattern.compile("^.*\\.") : null;

    /**
     * @return true on Java-5 or later; false on Java 1.4.x
     */
    public static boolean supportsAnnotations() {
        return annotationsAreSupported;
    }

    /**
     * On Java-5 or later the method tries to wrap a class with the same
     * name as the argument class but from the "annotation" sub-package.
     * On Java 1.4.x the method will simply wrap the argument class.
     *
     * @param c a class that is byte-code compatible with Java 1.4.x
     */
    public static ClassWrapper compatibleImplementationOf(Class c) {
        if (supportsAnnotations()) {
            try {
                c = Class.forName(packageName
                        .matcher(c.getName()).replaceFirst("$0annotation."));
            } catch (ClassNotFoundException ex) {
                throw ExceptionUtil.unchecked(ex);
            }
        }
        return new ClassWrapper(c);
    }
}
