/*
 * Copyright 2010 the original author or authors.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.callbackparams.support;

import org.apache.bcel.classfile.Method;
import org.apache.bcel.generic.ClassGen;
import org.apache.bcel.generic.Type;

/**
 * Immutable hash-key implementation that can be used as method-id, regardless
 * of whether the method is originally references through an instance of
 * "java.lang.reflect.Method" or "org.apache.bcel.classfile.Method".
 * The method is identified by class name, method name, return-type name and the
 * names of the argument types. (The return-type is part of the method id so
 * that the impact of generics will be handled properly.)
 *
 * @author Henrik Kaipe
 */
public class MethodHashKey {

    private final String className;
    final String methodName; // Not private as it will be accessed within package
    private final String signature;

    public MethodHashKey(String className, String methodName, String signature){
        this.className = className;
        this.methodName = methodName;
        this.signature = signature;
    }

    public static MethodHashKey getHashKey(java.lang.reflect.Method m) {
        return new MethodHashKey(m.getDeclaringClass().getName(),
                m.getName(), Type.getSignature(m));
    }

    public static MethodHashKey getHashKey(ClassGen cg, Method m) {
        return new MethodHashKey(cg.getClassName(),
                m.getName(), m.getSignature());
    }

    public boolean equals(Object obj) {
        if (obj == null) {
            return false;
        }
        if (getClass() != obj.getClass()) {
            return false;
        }
        final MethodHashKey other = (MethodHashKey) obj;
        if ((this.className == null) ? (other.className != null) : !this.className.equals(other.className)) {
            return false;
        }
        if ((this.methodName == null) ? (other.methodName != null) : !this.methodName.equals(other.methodName)) {
            return false;
        }
        if ((this.signature == null) ? (other.signature != null) : !this.signature.equals(other.signature)) {
            return false;
        }
        return true;
    }

    public int hashCode() {
        int hash = 3;
        hash = 53 * hash + (this.className != null ? this.className.hashCode() : 0);
        hash = 53 * hash + (this.methodName != null ? this.methodName.hashCode() : 0);
        hash = 53 * hash + (this.signature != null ? this.signature.hashCode() : 0);
        return hash;
    }
}
