/*
 * Copyright 2010-2013 the original author or authors.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.callbackparams.sandbox;

import java.beans.IntrospectionException;
import java.beans.Introspector;
import java.beans.PropertyDescriptor;
import java.beans.PropertyEditor;
import java.beans.PropertyEditorManager;
import java.lang.annotation.ElementType;
import java.lang.annotation.Retention;
import java.lang.annotation.RetentionPolicy;
import java.lang.annotation.Target;
import java.lang.reflect.Method;
import org.callbackparams.CallbackFactory;
import org.callbackparams.ext.WrappingCallbackImpls;
import org.callbackparams.wrap.WrappingSupport;

/**
 * This class has nested resources that can be used to produce callback-values
 * for {@link Property.Callback} through annotations.
 * <br/><br/>
 * Example usage ...
 * <pre>
 * &#64;RunWith({@link org.callbackparams.junit4.CallbackParamsRunner})
 * public class MyTest {
 *
 *   &#64;Test
 *   public void runTest(Property.Callback&lt;MyBean&gt; beanCallback) {
 *     // ...
 *   }
 *
 *   <b>&#64;Property.BeanClass(MyBean.class)
 *   &#64;Property.Name("myIntProperty")</b>
 *   enum MyEnum {
 *       <b>&#64;Property.Value("21")</b> FOO,
 *       <b>&#64;Property.Value("67")</b> BAR;
 *   }
 * }
 * </pre>
 * The annotations around and within <code>MyEnum</code> are meant as a shortcut
 * that is equivalent with this ...
 * <pre>
 *   enum MyEnum <b>implements Property.Callback&lt;MyBean&gt;</b> {
 *     FOO<b>(21)</b>, BAR<b>(67)</b>;
 *<b>
 *     int value;
 *
 *     MyEnum(int value) {
 *       this.value = value;
 *     }
 *
 *     void setValues(MyBean bean) {
 *         bean.setMyIntProperty(value);
 *     }
 *
 *     void assertValues(MyBean bean) {
 *       Assert.assertEquals("Value for property \"myIntProperty\" on " + bean,
 *           value, bean.getMyIntProperty());
 *     }</b>
 *   }
 * </pre>
 * Thus, a considerable amount of (<b>bold</b>) boilerplate code
 * can be excluded by using the
 * annotations, as shown in the first case.
 * <br/><br/>
 * This class is meant as a proof-of-concept for the Wrapping API that can
 * be used to specify an implementation of a callback-interface from a
 * callback-value that neither implements the callback-interface nor provides
 * any implementation
 * through {@link org.callbackparams.CallbackFactory#getCallback()}.
 * <br/>
 * This class being a proof-of-concept means that it applies to a
 * concept that is generally understood among developers (i.e. beans and
 * their properties) but unlike the other sandbox utility {@link Collector}
 * the reusability of this class is probably more limited.
 *
 * @author Henrik Kaipe
 */
public class Property {

    @Retention(RetentionPolicy.RUNTIME)
    @Target({ElementType.FIELD, ElementType.TYPE})
    public @interface BeanClass {
        java.lang.Class value();
    }

    @Retention(RetentionPolicy.RUNTIME)
    @Target({ElementType.FIELD, ElementType.TYPE})
    public @interface Name {
        String value();
    }

    @Retention(RetentionPolicy.RUNTIME)
    @Target({ElementType.FIELD, ElementType.TYPE})
    @WrappingCallbackImpls(BasicImpl.class)
    public @interface Value {
        String value();
    }

    public interface Callback<T> {
        void setValues(T bean);
        void assertValues(T bean);
    }

    protected static class BasicImpl<T> extends WrappingSupport<T>
    implements CallbackFactory, Callback<T> {

        /*
         * Annotation-values that will be injected ...
         */
        protected BeanClass beanClass;
        protected Name propertyName;
        protected Value propertyValue;

        /*
         * Derived values ...
         */
        final PropertyDescriptor descriptor = lookupPropertyDescriptor();
        final Object[] setterArgument = {createSetterArgument()};

        /**
         * To declare checked exceptions during initialization ...
         */
        BasicImpl() throws IntrospectionException {}

        private PropertyDescriptor lookupPropertyDescriptor()
        throws IntrospectionException {
            if (null != beanClass && null != propertyName) {
                for (PropertyDescriptor pd : Introspector
                        .getBeanInfo(beanClass.value())
                        .getPropertyDescriptors()) {
                    if (pd.getName().equals(propertyName.value())) {
                        return pd;
                    }
                }
            }
            return null;
        }

        private Object createSetterArgument() {
            if (null == descriptor) {
                return null;
            }
            PropertyEditor editor = PropertyEditorManager
                    .findEditor(descriptor.getPropertyType());
            editor.setAsText(propertyValue.value());
            return editor.getValue();
        }

        private static Object invoke(Method m, Object bean, Object[] args) {
            try {
                try {
                    m.setAccessible(true);
                } catch (SecurityException ignore) {
                    /* Never mind */
                }
                return m.invoke(bean, args);

            } catch (Exception x) {
                /*ExceptionUtil is not an API-class! so don't do this at home:*/
                throw org.callbackparams.support.ExceptionUtil.unchecked(x);
            }
        }

        /**
         * If all of the necessary annotations
         * {@link #beanClass}, {@link #propertyName} and {@link #propertyValue}
         * are found on the {@link #wrappedValue() wrapped value} (or its type)
         * then the method will return itself; otherwise null is returned.
         *
         * @return this object - if all of the necessary annotations
         * {@link #beanClass}, {@link #propertyName} and {@link #propertyValue}
         * are found on the {@link #wrappedValue() wrapped value} (or its type);
         * otherwise null
         */
        public Callback<T> getCallback() {
            if (null == propertyValue) {
                throw new IllegalStateException("Wrapped value - "
                        + wrappedValue()
                        + " - does not have annotation '@Value'");

            } else if (null == propertyName) {
                System.err.println(wrappedValue()
                        + " does not have annotation '@Name'");
                return null;

            } else if (null == beanClass) {
                System.err.println(wrappedValue()
                        + " does not have annotation '@BeanClass'");
                return null;
            }
            return this;
        }

        public void setValues(Object bean) {
            if (beanClass.value().isInstance(bean)) {
                invoke(descriptor.getWriteMethod(), bean, setterArgument);
            }
        }

        public void assertValues(Object bean) {
            if (beanClass.value().isInstance(bean)) {
                Object actual = invoke(descriptor.getReadMethod(), bean, null);
                if (false == setterArgument[0].equals(actual)) {
                    throw new AssertionError("Value for property \""
                            + descriptor.getName() + "\" on " + bean
                            + ": expected <" + setterArgument[0]
                            + "> but was <" + actual + ">");
                }
            }
        }
    }

    private Property() {}
}
